function compute_U(s,order,opt,Img00,Img01,dirname,kappa)
% estimate displacement with SSD or modified SSD criterion
% s : half-size of the subset
% order : order of basis function  (0, 1 or 2)
% opt : '1D' or '2D' correction, 'GT' for "true" coef with known GT displacement, 'NO' no correction
% Img00 : reference state
% Img01 : deformed state
% dirname: output directory to save figures
% kappa: regularization parameter in modified SSD criteria

% ground truth displacement (Equation 4.4)
[x,y]=meshgrid(1:size(Img00,2),1:size(Img00,1));
amplitude_def=.5;
period_min=5;
period_max=50;
def_y=amplitude_def*cos(2*pi*(y-50)./((period_max-period_min).*x/500+period_min));
def_x=zeros(size(x));

m=floor((size(Img00,1)-1)/2);
ss=s+2;  % slightly larger window to prevent border effects, but SSD criterion is actually computed on a subset of half-size s (cf SSD.m)

% n : number of displacement parameters
switch order
    case 0 % constant
        n=2;
    case 1 % affine
        n=6;
    case 2 % quadratic
        n=12;
end

% interpolation scheme, 'cubic' in Section 4
interp_scheme = 'cubic'

retrieved_displacement=zeros(size(Img00,1),size(Img00,2),n);
predicted_displacement=zeros(size(Img00,1),size(Img00,2),n);
corrected_displacement=zeros(size(Img00,1),size(Img00,2),n);

% options of fminunc, to prevent early stopping
options=optimoptions('fminunc','Display','off','MaxIterations',10000,'OptimalityTolerance',1e-10,'StepTolerance',1e-10);

h = waitbar(0,'Initializing waitbar...','Name','SSD minimization');
for i=1+ss:size(Img00,1)-ss
    ratio=100*(i-ss)/(size(Img00,1)-2*ss);
    waitbar(ratio/100,h,sprintf('%d%% along...',uint8(ratio))) 
    parfor j=1+ss:size(Img00,2)-ss
        Ima_ref=Img00(i-ss:i+ss,j-ss:j+ss);
        Ima_def=Img01(i-ss:i+ss,j-ss:j+ss);
        fSSD = @(lambda)SSD(lambda,Ima_def,Ima_ref,interp_scheme,opt);
        lambda=fminunc(fSSD,zeros(n,1),options);
        retrieved_displacement(i,j,:)=lambda;  % retrieved global displacement is the local parametric displacement at the center of the subset
        
        % calculate L and G to predict lambda (Theorem 2.1 in Section 2.1)
        loc_def_x=def_x(i-ss:i+ss,j-ss:j+ss);  % GT displacement
        loc_def_y=def_y(i-ss:i+ss,j-ss:j+ss);
        [L,G] = calculate_LandG(length(lambda),Ima_def,loc_def_x,loc_def_y,interp_scheme);
        if (strcmp(opt,'1D')) L=L(:,2:2:end); end  % if 1D-displacement (with 1D-SSD criterion)
        lambda_pred=(L'*L)\L'*G;
        if (strcmp(opt,'1D'))  % if 1D-displacement (with 1D-SSD criterion) 
            lambda_pred2=zeros(1,2*length(lambda_pred)); lambda_pred2(:,2:2:end)=lambda_pred; lambda_pred=lambda_pred2; 
        end
        predicted_displacement(i,j,:)=lambda_pred;
    end
end
close(h)


% modified SSD criteria (Section 4.2.5)
if (~strcmp(opt,'NO'))
    h = waitbar(0,'Initializing waitbar...','Name','SSD tilde minimization');
    for i=1+ss:size(Img00,1)-ss
        ratio=100*(i-ss)/(size(Img00,1)-2*ss);
        waitbar(ratio/100,h,sprintf('%d%% along...',uint8(ratio))) 
        parfor j=1+ss:size(Img00,2)-ss
            lambda_init=retrieved_displacement(i,j,:);
            Ima_ref=Img00(i-ss:i+ss,j-ss:j+ss);
            Ima_def=Img01(i-ss:i+ss,j-ss:j+ss);
            loc_def_x=def_x(i-ss:i+ss,j-ss:j+ss);
            loc_def_y=def_y(i-ss:i+ss,j-ss:j+ss);
            fSSDtilde = @(lambda)SSDtilde(lambda,Ima_def,Ima_ref,loc_def_x,loc_def_y,'cubic',opt,kappa);  
            lambda=fminunc(fSSDtilde,lambda_init,options);
            corrected_displacement(i,j,:)=lambda;
        end
    end
    close(h)
end


% generate and save figures in "dirname" directory

t=2*s+1;  % subset size

sgf=zeros(size(def_x,1),size(def_x,2),2);
def=zeros(size(def_x,1),size(def_x,2),2);

% 2D SG filter comes from:
% https://fr.mathworks.com/matlabcentral/fileexchange/37147-savitzky-golay-smoothing-filter-for-2d-data
sgf(:,:,1) = savitzkyGolay2D_rle_coupling(size(def_x,2),size(def_x,1),def_x,t,t,order);
sgf(:,:,2) = savitzkyGolay2D_rle_coupling(size(def_y,2),size(def_y,1),def_y,t,t,order);
def(:,:,1) = def_x;
def(:,:,2) = def_y;

cmap=[0, 0.4470, 0.7410; 0.8500, 0.3250, 0.0980; 0.9290, 0.6940, 0.1250; 0.4940, 0.1840, 0.5560; 0.4660, 0.6740, 0.1880; 0.3010, 0.7450, 0.9330; 0.6350, 0.0780, 0.1840];

if (~strcmp(opt,'1D'))
maxval=quantile(max(abs(retrieved_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,1))),.7);
if (maxval==0) maxval=.1; end
scale_x=[-maxval,maxval];
hfig=figure;
set(hfig, 'Position', [100 100 560 800])
subplot(5,1,1)
imagesc(ss,ss,retrieved_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,1),scale_x), colorbar, axis tight
hold on
line([ss,500-ss], [50,50],'LineWidth',2,'Color',cmap(1,:));
ti=title('retrieved with SSD');
set(ti,'Fontweight','normal','Fontsize',12)
subplot(5,1,2)
imagesc(ss,ss,predicted_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,1),scale_x), colorbar, axis tight
hold on
line([ss,500-ss], [50,50],'LineStyle','--','LineWidth',2,'Color',cmap(2,:));
ti=title('predicted');
set(ti,'Fontweight','normal','Fontsize',12)
ind=3;
if (~strcmp(opt,'NO'))
    subplot(5,1,3)
    imagesc(ss,ss,corrected_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,1),scale_x), colorbar, axis tight
    hold on
    line([ss,500-ss], [50,50],'LineWidth',2,'Color',cmap(5,:));
    ti=title('retrieved with $\widetilde{\mathrm{SSD}}$');
    set(ti,'Interpreter','latex','Fontweight','normal','Fontsize',12)
    ind=ind+1;
end
subplot(5,1,ind)
imagesc(ss,ss,sgf(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,1),scale_x), colorbar, axis tight
hold on
line([ss,500-ss], [50,50],'LineStyle','--','LineWidth',2,'Color',cmap(3,:));
ti=title('output of SG filter');
set(ti,'Fontweight','normal','Fontsize',12)
ind=ind+1;
subplot(5,1,ind)
imagesc(ss,ss,def(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,1),scale_x), colorbar, axis  tight
hold on
line([ss,500-ss], [50,50],'LineWidth',2,'Color',cmap(4,:));
ti=title('GT displacement');
set(ti,'Fontweight','normal','Fontsize',12)
saveas(hfig,strcat(dirname,'fig_dispx_',num2str(s),'_',num2str(order),'_',opt,'.pdf'));
end

maxval=max(max(abs(def(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,2))));
if (maxval==0) maxval=.1; end
scale_y=[-maxval,maxval];
hfig=figure;
set(hfig, 'Position', [150 150 560 800])
subplot(5,1,1)
imagesc(ss,ss,retrieved_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,2),scale_y), colorbar, axis  tight
hold on
line([ss,500-ss], [50,50],'LineWidth',2,'Color',cmap(1,:));
ti=title('retrieved with SSD');
set(ti,'Fontweight','normal','Fontsize',12)
subplot(5,1,2)
imagesc(ss,ss,predicted_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,2),scale_y), colorbar, axis  tight
hold on
line([ss,500-ss], [50,50],'LineStyle','--','LineWidth',2,'Color',cmap(2,:));
ti=title('predicted');
set(ti,'Fontweight','normal','Fontsize',12)
ind=3;
if (~strcmp(opt,'NO'))
    subplot(5,1,3)
    imagesc(ss,ss,corrected_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,2),scale_y), colorbar, axis  tight
    hold on
    line([ss,500-ss], [50,50],'LineWidth',2,'Color',cmap(5,:));
    ti=title('retrieved with $\widetilde{\mathrm{SSD}}$');
    set(ti,'Interpreter','latex','Fontweight','normal','Fontsize',12)
    ind=ind+1;
end
subplot(5,1,ind)
imagesc(ss,ss,sgf(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,2),scale_y), colorbar, axis  tight
hold on
line([ss,500-ss], [50,50],'LineStyle','--','LineWidth',2,'Color',cmap(3,:));
ti=title('output of SG filter');
set(ti,'Fontweight','normal','Fontsize',12)
ind=ind+1;
subplot(5,1,ind)
imagesc(ss,ss,def(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,2),scale_y), colorbar, axis  tight
hold on
line([s,500-ss], [50,50],'LineWidth',2,'Color',cmap(4,:));
ti=title('GT displacement');
set(ti,'Fontweight','normal','Fontsize',12)
saveas(hfig,strcat(dirname,'fig_dispy_',num2str(s),'_',num2str(order),'_',opt,'.pdf'));


diff_retrieved_displacement=retrieved_displacement(:,:,1:2)-sgf;
diff_predicted_displacement=predicted_displacement(:,:,1:2)-sgf;
if (~strcmp(opt,'NO')) diff_corrected_displacement=corrected_displacement(:,:,1:2)-sgf; end

if (~strcmp(opt,'1D'))
maxval=max(max(abs(diff_retrieved_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,1))))*.7;
if (maxval==0) maxval=.1; end
scale_diff=[-maxval,maxval];
hfig=figure;
subplot(2,1,1)
imagesc(ss,ss,diff_retrieved_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,1),scale_diff), colorbar, axis  tight
ti=title('retrieved with SSD - output of SG filter');
set(ti,'Fontweight','normal','Fontsize',12)
if (~strcmp(opt,'NO'))
    subplot(2,1,2)
    imagesc(ss,ss,diff_corrected_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,1),scale_diff), colorbar, axis  tight
    ti=title('retrieved with $\widetilde{\mathrm{SSD}}$ - output of SG filter');
    set(ti,'Interpreter','latex','Fontweight','normal','Fontsize',12)
end
saveas(hfig,strcat(dirname,'fig_diffx_',num2str(s),'_',num2str(order),'_',opt,'.pdf'));
end

maxval=max(max(abs(diff_retrieved_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,2))))*.7;
if (maxval==0) maxval=.1; end
scale_diff=[-maxval,maxval];
hfig=figure;
subplot(2,1,1)
imagesc(ss,ss,diff_retrieved_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,2),scale_diff), colorbar, axis  tight
ti=title('retrieved with SSD - output of SG filter');
set(ti,'Fontweight','normal','Fontsize',12)
if (~strcmp(opt,'NO'))
    subplot(2,1,2)
    imagesc(ss,ss,diff_corrected_displacement(1+ss:size(Img00,1)-ss,1+ss:size(Img00,2)-ss,2),scale_diff), colorbar, axis  tight
    ti=title('retrieved with $\widetilde{\mathrm{SSD}}$ - output of SG filter');
    set(ti,'Interpreter','latex','Fontweight','normal','Fontsize',14)
end
saveas(hfig,strcat(dirname,'fig_diffy_',num2str(s),'_',num2str(order),'_',opt,'.pdf'));

N=(size(Img00,1)-2*ss)*(size(Img00,2)-2*ss);

for k=1:2
    if ((strcmp(opt,'1D'))&(k==1)) continue; end
    if (k==1) caption='$\phi_x=\lambda_1$'; else caption='$\phi_y=\lambda_2$'; end
    hfig=figure;
    if (~strcmp(opt,'NO'))
        h=plot(1+ss:size(Img00,2)-ss,retrieved_displacement(m,1+ss:size(Img00,2)-ss,k),...
            1+ss:size(Img00,2)-ss,predicted_displacement(m,1+ss:size(Img00,2)-ss,k),'--',...
            1+ss:size(Img00,2)-ss,sgf(m,1+ss:size(Img00,2)-ss,k),'--',...
            1+ss:size(Img00,2)-ss,def(m,1+ss:size(Img00,2)-ss,k),...
            1+ss:size(Img00,2)-ss,corrected_displacement(m,1+ss:size(Img00,2)-ss,k),'LineWidth',2);
        hold on, plot(NaN,NaN,'linestyle', 'none');
        grid on
        lgd=legend('retrieved with SSD','predicted', 'output of SG filter','GT displacement','retrieved with $\widetilde{\mathrm{SSD}}$',caption);
        set(lgd,'fontsize',11,'Interpreter','latex','Location','southeast')
    else
        h=plot(1+ss:size(Img00,2)-ss,retrieved_displacement(m,1+ss:size(Img00,2)-ss,k),...
            1+ss:size(Img00,2)-ss,predicted_displacement(m,1+ss:size(Img00,2)-ss,k),'--',...
            1+ss:size(Img00,2)-ss,sgf(m,1+ss:size(Img00,2)-ss,k),'--',...
            1+ss:size(Img00,2)-ss,def(m,1+ss:size(Img00,2)-ss,k),'LineWidth',2);
        hold on, plot(NaN,NaN,'linestyle', 'none');
        grid on
        lgd=legend('retrieved with SSD','predicted', 'output of SG filter','GT displacement',caption);
        set(lgd,'fontsize',11,'Interpreter','latex','Location','southeast')
    end
    saveas(hfig,strcat(dirname,'fig_plot_',num2str(k),'_',num2str(s),'_',num2str(order),'_',opt,'.pdf'));
end


if (order>0)
    % function sgdf_2d comes from:
    % https://fr.mathworks.com/matlabcentral/fileexchange/6151-2-d-savitzky-golay-differentiation-filter
    SGF=sgdf_2d(-s:s,-s:s,1,1,1);
    def_deriv_SGF=zeros(size(def,1),size(def,2),4);
    def_deriv_SGF(:,:,1)=imfilter(def(:,:,1),SGF,'same');
    def_deriv_SGF(:,:,2)=imfilter(def(:,:,2),SGF,'same');
    def_deriv_SGF(:,:,3)=imfilter(def(:,:,1),SGF','same');
    def_deriv_SGF(:,:,4)=imfilter(def(:,:,2),SGF','same');
    for k=3:n
        if ((strcmp(opt,'1D'))&(mod(k,2)==1)) continue; end
        hfig=figure;
        caption=strcat('$\lambda_{',num2str(k),'}$');
        if (n<7) 
            if (~strcmp(opt,'NO'))
                h=plot(1+ss:size(Img00,2)-ss,retrieved_displacement(m,1+ss:size(Img00,2)-ss,k),...
                1+ss:size(Img00,2)-ss,predicted_displacement(m,1+ss:size(Img00,2)-ss,k),'--',...
                1+ss:size(Img00,2)-ss,def_deriv_SGF(m,1+ss:size(Img00,2)-ss,k-2),'--',...
                1+ss:size(Img00,2)-ss,corrected_displacement(m,1+ss:size(Img00,2)-ss,k),'LineWidth',2);
                hold on, plot(NaN,NaN,'linestyle', 'none');
                lgd=legend('retrieved with SSD','predicted', 'output of SG filter','retrieved with $\widetilde{\mathrm{SSD}}$',caption);
                set(h(4),'Color',[0.4660, 0.6740, 0.1880]);
            else
                h=plot(1+ss:size(Img00,2)-ss,retrieved_displacement(m,1+ss:size(Img00,2)-ss,k),...
                1+ss:size(Img00,2)-ss,predicted_displacement(m,1+ss:size(Img00,2)-ss,k),'--',...
                1+ss:size(Img00,2)-ss,def_deriv_SGF(m,1+ss:size(Img00,2)-ss,k-2),'--','LineWidth',2);
                hold on, plot(NaN,NaN,'linestyle', 'none');
                lgd=legend('retrieved with SSD','predicted', 'output of SG filter',caption);
            end           
        else
            if (~strcmp(opt,'NO'))
                h=plot(1+ss:size(Img00,2)-ss,retrieved_displacement(m,1+ss:size(Img00,2)-ss,k),...
                1+ss:size(Img00,2)-ss,predicted_displacement(m,1+ss:size(Img00,2)-ss,k),'--',...
                1+ss:size(Img00,2)-ss,corrected_displacement(m,1+ss:size(Img00,2)-ss,k),'LineWidth',2);
                set(h(3),'Color',[0.4660, 0.6740, 0.1880]);
                hold on, plot(NaN,NaN,'linestyle', 'none');
                lgd=legend('retrieved with SSD','predicted','retrieved with $\widetilde{\mathrm{SSD}}$',caption);
            else
                h=plot(1+ss:size(Img00,2)-ss,retrieved_displacement(m,1+ss:size(Img00,2)-ss,k),...
                1+ss:size(Img00,2)-ss,predicted_displacement(m,1+ss:size(Img00,2)-ss,k),'--','LineWidth',2);
                hold on, plot(NaN,NaN,'linestyle', 'none');
                lgd=legend('retrieved with SSD','predicted',caption);
            end
        end
        grid on
        set(lgd,'fontsize',11,'Interpreter','latex','Location','southeast')
        saveas(hfig,strcat(dirname,'fig_plot_',num2str(k),'_',num2str(s),'_',num2str(order),'_',opt,'.pdf'));
    end
end

